﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Problems.MultiObjectiveTestFunctions.Comparators;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  /// <summary>
  /// The Hyprevolume-metric is defined as the Hypervolume enclosed between a given reference point,
  /// that is fixed for every evaluation function and the evaluated front.
  /// 
  /// Example:
  /// r is the reference Point at (1|1)  and every Point p is part of the evaluated front
  /// The filled Area labled HV is the 2 diensional Hypervolume enclosed by this front. 
  /// 
  /// (0|1)                (1|1)
  ///   +      +-------------r
  ///   |      |###### HV ###|
  ///   |      p------+######|
  ///   |             p+#####|
  ///   |              |#####|
  ///   |              p-+###|
  ///   |                p---+
  ///   |                 
  ///   +--------------------1
  /// (0|0)                (1|0)
  /// 
  ///  Please note that in this example both dimensions are minimized. The reference Point need to be dominated by EVERY point in the evaluated front 
  /// 
  /// </summary>
  public class Hypervolume : IMultiObjectiveDistance {

    private double[] reference;
    private bool[] maximization;
    public Hypervolume(double[] reference, bool[] maximization) {
      if (reference.Length != 2) throw new Exception("Only 2-dimensional cases are supported yet");
      this.reference = reference;
      this.maximization = maximization;
    }

    public double Compare(IEnumerable<double[]> front, IEnumerable<double[]> optimalFront) {
      return GetHypervolume(optimalFront) - GetHypervolume(front);

    }

    public double GetHypervolume(IEnumerable<double[]> front) {
      if (front == null) throw new ArgumentException("Fronts must not be null");
      //TODO what to do if set contains dominated points
      double[][] set = front.ToArray();   //Still no Good
      if (set.Length == 0) throw new ArgumentException("Fronts must not be empty");
      Array.Sort<double[]>(set, Utilities.getDimensionComparer(0, maximization[0]));
      double[] last = set[set.Length - 1];
      CheckConsistency(last, 0);
      CheckConsistency(last, 1);
      double sum = 0;
      for (int i = 0; i < set.Length - 1; i++) {
        CheckConsistency(set[i], 1);
        sum += Math.Abs((set[i][0] - set[i + 1][0])) * Math.Abs((set[i][1] - reference[1]));
      }

      sum += Math.Abs(reference[0] - last[0]) * Math.Abs(reference[1] - last[1]);
      return sum;
    }

    public static double GetHypervolume(IEnumerable<double[]> front, double[] reference, bool[] maximization){
      Hypervolume comp = new Hypervolume(reference, maximization);
      return comp.GetHypervolume(front);
    }

    public static double GetDistance(IEnumerable<double[]> front, IEnumerable<double[]> optimalFront, double[] reference, bool[] maximization) {
      return GetHypervolume(optimalFront, reference, maximization) - GetHypervolume(front, reference, maximization);
    }

    private void CheckConsistency(double[] point, int dim) {
      if (!maximization[dim] && point[dim] > reference[dim]) throw new ArgumentException("Reference Point must be dominated by all points of the front");
      if (maximization[dim] && point[dim] < reference[dim]) throw new ArgumentException("Reference Point must be dominated by all points of the front");
      if (point.Length != 2) throw new ArgumentException("Only 2-dimensional cases are supported yet");
    }
  }
}
